package ru.dbotthepony.kstarbound.client

import org.lwjgl.glfw.GLFW.glfwGetTime
import ru.dbotthepony.kstarbound.api.IStruct2d
import ru.dbotthepony.kstarbound.api.IStruct2f
import ru.dbotthepony.kstarbound.client.render.ChunkRenderer
import ru.dbotthepony.kstarbound.client.render.renderLayeredList
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.world.*

class ClientWorldChunkTuple(
	world: World<*>,
	chunk: Chunk,
	top: IWorldChunkTuple?,
	left: IWorldChunkTuple?,
	right: IWorldChunkTuple?,
	bottom: IWorldChunkTuple?,

	val renderer: ChunkRenderer
) : MutableWorldChunkTuple(
	world,
	chunk,
	top,
	left,
	right,
	bottom,
)

class ClientWorld(val client: StarboundClient, seed: Long = 0L) : World<ClientWorldChunkTuple>(seed) {
	override fun tupleFactory(
		chunk: Chunk,
		top: IWorldChunkTuple?,
		left: IWorldChunkTuple?,
		right: IWorldChunkTuple?,
		bottom: IWorldChunkTuple?
	): ClientWorldChunkTuple {
		return ClientWorldChunkTuple(
			world = this,
			chunk = chunk,
			top = top,
			left = left,
			right = right,
			bottom = bottom,

			renderer = ChunkRenderer(client.gl, chunk, this)
		)
	}

	/**
	 * Отрисовывает этот с обрезкой невидимой геометрии с точки зрения [size] в Starbound Units
	 *
	 * Все координаты "местности" сохраняются, поэтому, если отрисовывать слишком далеко от 0, 0
	 * то геометрия может начать искажаться из-за погрешности плавающей запятой
	 *
	 * Обрезает всю заведомо невидимую геометрию на основе аргументов mins и maxs (в пикселях)
	 */
	fun render(
		size: AABB,
	) {
		val determineRenderers = ArrayList<ChunkRenderer>()

		for (chunk in collectInternal(size.encasingChunkPosAABB())) {
			determineRenderers.add(chunk.renderer)
		}

		for (renderer in determineRenderers) {
			renderer.autoBakeStatic()
		}

		renderLayeredList(client.gl.matrixStack, determineRenderers)

		for (renderer in determineRenderers) {
			renderer.renderDebug()
		}
	}

	override fun thinkInner(delta: Double) {
		for (ent in entities) {
			ent.think(delta)
		}
	}
}
